@@ -2,36 +2,41 @@ use crate::PyProjectToml;
 use anyhow::{bail, format_err, Context, Result};
 use fs_err as fs;
 use indexmap::IndexMap;
+use normpath::PathExt;
 use pep440_rs::{Version, VersionSpecifiers};
-use pep508_rs::{MarkerExpression, MarkerOperator, MarkerTree, MarkerValue, Requirement};
-use pyproject_toml::License;
+use pep508_rs::{
+    ExtraName, ExtraOperator, MarkerExpression, MarkerTree, MarkerValueExtra, Requirement,
+};
+use pyproject_toml::{check_pep639_glob, License};
 use regex::Regex;
 use serde::{Deserialize, Serialize};
-use std::collections::HashMap;
+use std::collections::{HashMap, HashSet};
 use std::fmt::Write as _;
 use std::path::{Path, PathBuf};
 use std::str;
 use std::str::FromStr;
+use tracing::debug;
 
 /// The metadata required to generate the .dist-info directory
 #[derive(Serialize, Deserialize, Debug, Clone, Eq, PartialEq)]
 pub struct WheelMetadata {
-    /// Python Package Metadata 2.1
-    pub metadata21: Metadata21,
+    /// Python Package Metadata 2.4
+    pub metadata24: Metadata24,
     /// The `[console_scripts]` for the entry_points.txt
     pub scripts: HashMap<String, String>,
     /// The name of the module can be distinct from the package name, mostly
     /// because package names normally contain minuses while module names
-    /// have underscores. The package name is part of metadata21
+    /// have underscores. The package name is part of metadata24
     pub module_name: String,
 }
 
-/// Python Package Metadata 2.1 as specified in
+/// Python Package Metadata 2.4 as specified in
 /// https://packaging.python.org/specifications/core-metadata/
+/// Maturin writes static metadata and does not support dynamic fields atm.
 #[derive(Serialize, Deserialize, Debug, Clone, Eq, PartialEq)]
 #[serde(rename_all = "kebab-case")]
 #[allow(missing_docs)]
-pub struct Metadata21 {
+pub struct Metadata24 {
     // Mandatory fields
     pub metadata_version: String,
     pub name: String,
@@ -50,7 +55,7 @@ pub struct Metadata21 {
     pub maintainer: Option<String>,
     pub maintainer_email: Option<String>,
     pub license: Option<String>,
-    // https://peps.python.org/pep-0639/#license-file-multiple-use
+    pub license_expression: Option<String>,
     pub license_files: Vec<PathBuf>,
     pub classifiers: Vec<String>,
     pub requires_dist: Vec<Requirement>,
@@ -65,11 +70,11 @@ pub struct Metadata21 {
     pub entry_points: IndexMap<String, IndexMap<String, String>>,
 }
 
-impl Metadata21 {
+impl Metadata24 {
     /// Initializes with name, version and otherwise the defaults
     pub fn new(name: String, version: Version) -> Self {
         Self {
-            metadata_version: "2.1".to_string(),
+            metadata_version: "2.4".to_string(),
             name,
             version,
             platform: vec![],
@@ -85,6 +90,7 @@ impl Metadata21 {
             maintainer: None,
             maintainer_email: None,
             license: None,
+            license_expression: None,
             license_files: vec![],
             classifiers: vec![],
             requires_dist: vec![],
@@ -122,7 +128,7 @@ fn path_to_content_type(path: &Path) -> String {
         })
 }
 
-impl Metadata21 {
+impl Metadata24 {
     /// Merge metadata with pyproject.toml, where pyproject.toml takes precedence
     ///
     /// pyproject_dir must be the directory containing pyproject.toml
@@ -133,7 +139,60 @@ impl Metadata21 {
     ) -> Result<()> {
         let pyproject_dir = pyproject_dir.as_ref();
         if let Some(project) = &pyproject_toml.project {
-            self.name = project.name.clone();
+            let dynamic: HashSet<&str> = project
+                .dynamic
+                .as_ref()
+                .map(|x| x.iter().map(AsRef::as_ref).collect())
+                .unwrap_or_default();
+            if dynamic.contains("name") {
+                bail!("`project.dynamic` must not specify `name` in pyproject.toml");
+            }
+
+            // According to PEP 621, build backends must not add metadata fields
+            // that are not declared in the dynamic list. Clear fields from Cargo.toml
+            // that are not in the dynamic list.
+            if !dynamic.contains("description") {
+                self.summary = None;
+            }
+            if !dynamic.contains("authors") {
+                self.author = None;
+                self.author_email = None;
+            }
+            if !dynamic.contains("maintainers") {
+                self.maintainer = None;
+                self.maintainer_email = None;
+            }
+            if !dynamic.contains("keywords") {
+                self.keywords = None;
+            }
+            if !dynamic.contains("urls") {
+                self.project_url.clear();
+            }
+            if !dynamic.contains("license") {
+                self.license = None;
+                // Don't clear license_files as they may come from auto-discovery
+            }
+            if !dynamic.contains("classifiers") {
+                self.classifiers.clear();
+            }
+            if !dynamic.contains("readme") {
+                self.description = None;
+                self.description_content_type = None;
+            }
+            if !dynamic.contains("requires-python") {
+                self.requires_python = None;
+            }
+
+            self.name.clone_from(&project.name);
+
+            let version_ok = pyproject_toml.warn_invalid_version_info();
+            if !version_ok {
+                // This is a hard error for maturin>=2.0, see https://github.com/PyO3/maturin/issues/2416.
+                let current_major = env!("CARGO_PKG_VERSION_MAJOR").parse::<usize>().unwrap();
+                if current_major > 1 {
+                    bail!("Invalid version information in pyproject.toml.");
+                }
+            }
 
             if let Some(version) = &project.version {
                 self.version = version.clone();
@@ -164,58 +223,102 @@ impl Metadata21 {
                     if let Some(readme_path) = file {
                         let readme_path = pyproject_dir.join(readme_path);
                         let description = Some(fs::read_to_string(&readme_path).context(format!(
-                                "Failed to read readme specified in pyproject.toml, which should be at {}",
-                                readme_path.display()
-                            ))?);
+                            "Failed to read readme specified in pyproject.toml, which should be at {}",
+                            readme_path.display()
+                        ))?);
                         self.description = description;
                     }
                     if let Some(description) = text {
                         self.description = Some(description.clone());
                     }
-                    self.description_content_type = content_type.clone();
+                    self.description_content_type.clone_from(content_type);
                 }
                 None => {}
             }
 
+            // Make sure existing license files from `Cargo.toml` are relative to the project root
+            let license_files = std::mem::take(&mut self.license_files);
+            for license_file in license_files {
+                let license_path = license_file
+                    .strip_prefix(pyproject_dir)
+                    .with_context(|| {
+                        format!(
+                            "license file `{}` exists outside of the project root `{}`",
+                            license_file.display(),
+                            pyproject_dir.display()
+                        )
+                    })?
+                    .to_path_buf();
+                self.license_files.push(license_path);
+            }
+
             if let Some(requires_python) = &project.requires_python {
                 self.requires_python = Some(requires_python.clone());
             }
 
             if let Some(license) = &project.license {
                 match license {
-                    // TODO: switch to License-Expression core metadata, see https://peps.python.org/pep-0639/#add-license-expression-field
-                    License::String(license_expr) => self.license = Some(license_expr.clone()),
-                    License::Table { file, text } => match (file, text) {
-                        (Some(_), Some(_)) => {
-                            bail!("file and text fields of 'project.license' are mutually-exclusive, only one of them should be specified");
-                        }
-                        (Some(license_path), None) => {
-                            let license_path = pyproject_dir.join(license_path);
-                            self.license_files.push(license_path);
-                        }
-                        (None, Some(license_text)) => self.license = Some(license_text.clone()),
-                        (None, None) => {}
-                    },
+                    // PEP 639
+                    License::Spdx(license_expr) => {
+                        self.license_expression = Some(license_expr.clone())
+                    }
+                    // Deprecated by PEP 639
+                    License::File { file } => {
+                        self.license_files.push(file.to_path_buf());
+                    }
+                    License::Text { text } => self.license = Some(text.clone()),
                 }
             }
 
-            // Until PEP 639 is approved with metadata 2.3, we can assume a
-            // dynamic license-files (also awaiting full 2.2 metadata support)
-            // We're already emitting the License-Files metadata without issue.
-            // license-files.globs = ["LICEN[CS]E*", "COPYING*", "NOTICE*", "AUTHORS*"]
-            let license_include_targets = ["LICEN[CS]E*", "COPYING*", "NOTICE*", "AUTHORS*"];
-            let escaped_manifest_string = glob::Pattern::escape(pyproject_dir.to_str().unwrap());
-            let escaped_manifest_path = Path::new(&escaped_manifest_string);
-            for pattern in license_include_targets.iter() {
-                for license_path in
-                    glob::glob(&escaped_manifest_path.join(pattern).to_string_lossy())?
-                        .filter_map(Result::ok)
-                {
-                    // if the pyproject.toml specified the license file,
-                    // then we won't list it as automatically included
-                    if !self.license_files.contains(&license_path) {
-                        eprintln!("📦 Including license file \"{}\"", license_path.display());
-                        self.license_files.push(license_path);
+            // Handle PEP 639 license-files field
+            if let Some(license_files) = &project.license_files {
+                // Safe on Windows and Unix as neither forward nor backwards slashes are escaped.
+                let escaped_pyproject_dir =
+                    PathBuf::from(glob::Pattern::escape(pyproject_dir.to_str().unwrap()));
+                for license_glob in license_files {
+                    check_pep639_glob(license_glob)?;
+                    for license_path in
+                        glob::glob(&escaped_pyproject_dir.join(license_glob).to_string_lossy())?
+                    {
+                        let license_path = license_path?;
+                        if !license_path.is_file() {
+                            continue;
+                        }
+                        let license_path = license_path
+                            .strip_prefix(pyproject_dir)
+                            .expect("matched path starts with glob root")
+                            .to_path_buf();
+                        if !self.license_files.contains(&license_path) {
+                            debug!("Including license file `{}`", license_path.display());
+                            self.license_files.push(license_path);
+                        }
+                    }
+                }
+            } else {
+                // Auto-discovery of license files for backwards compatibility
+                // license-files.globs = ["LICEN[CS]E*", "COPYING*", "NOTICE*", "AUTHORS*"]
+                let license_include_targets = ["LICEN[CS]E*", "COPYING*", "NOTICE*", "AUTHORS*"];
+                let escaped_manifest_string =
+                    glob::Pattern::escape(pyproject_dir.to_str().unwrap());
+                let escaped_manifest_path = Path::new(&escaped_manifest_string);
+                for pattern in license_include_targets.iter() {
+                    for license_path in
+                        glob::glob(&escaped_manifest_path.join(pattern).to_string_lossy())?
+                            .filter_map(Result::ok)
+                    {
+                        if !license_path.is_file() {
+                            continue;
+                        }
+                        let license_path = license_path
+                            .strip_prefix(pyproject_dir)
+                            .expect("matched path starts with glob root")
+                            .to_path_buf();
+                        // if the pyproject.toml specified the license file,
+                        // then we won't list it as automatically included
+                        if !self.license_files.contains(&license_path) {
+                            eprintln!("📦 Including license file `{}`", license_path.display());
+                            self.license_files.push(license_path);
+                        }
                     }
                 }
             }
@@ -224,15 +327,15 @@ impl Metadata21 {
                 let mut names = Vec::with_capacity(authors.len());
                 let mut emails = Vec::with_capacity(authors.len());
                 for author in authors {
-                    match (&author.name, &author.email) {
+                    match (author.name(), author.email()) {
                         (Some(name), Some(email)) => {
-                            emails.push(format!("{name} <{email}>"));
+                            emails.push(escape_email_with_display_name(name, email));
                         }
                         (Some(name), None) => {
-                            names.push(name.as_str());
+                            names.push(name);
                         }
                         (None, Some(email)) => {
-                            emails.push(email.clone());
+                            emails.push(email.to_string());
                         }
                         (None, None) => {}
                     }
@@ -249,15 +352,15 @@ impl Metadata21 {
                 let mut names = Vec::with_capacity(maintainers.len());
                 let mut emails = Vec::with_capacity(maintainers.len());
                 for maintainer in maintainers {
-                    match (&maintainer.name, &maintainer.email) {
+                    match (maintainer.name(), maintainer.email()) {
                         (Some(name), Some(email)) => {
-                            emails.push(format!("{name} <{email}>"));
+                            emails.push(escape_email_with_display_name(name, email));
                         }
                         (Some(name), None) => {
-                            names.push(name.as_str());
+                            names.push(name);
                         }
                         (None, Some(email)) => {
-                            emails.push(email.clone());
+                            emails.push(email.to_string());
                         }
                         (None, None) => {}
                     }
@@ -275,15 +378,15 @@ impl Metadata21 {
             }
 
             if let Some(classifiers) = &project.classifiers {
-                self.classifiers = classifiers.clone();
+                self.classifiers.clone_from(classifiers);
             }
 
             if let Some(urls) = &project.urls {
-                self.project_url = urls.clone();
+                self.project_url.clone_from(urls);
             }
 
             if let Some(dependencies) = &project.dependencies {
-                self.requires_dist = dependencies.clone();
+                self.requires_dist.clone_from(dependencies);
             }
 
             if let Some(dependencies) = &project.optional_dependencies {
@@ -293,26 +396,24 @@ impl Metadata21 {
                     for dep in deps {
                         let mut dep = dep.clone();
                         // Keep in sync with `develop()`!
-                        let new_extra = MarkerTree::Expression(MarkerExpression {
-                            l_value: MarkerValue::Extra,
-                            operator: MarkerOperator::Equal,
-                            r_value: MarkerValue::QuotedString(extra.to_string()),
-                        });
-                        if let Some(existing) = dep.marker.take() {
-                            dep.marker = Some(MarkerTree::And(vec![existing, new_extra]));
-                        } else {
-                            dep.marker = Some(new_extra);
-                        }
+                        let new_extra = MarkerExpression::Extra {
+                            operator: ExtraOperator::Equal,
+                            name: MarkerValueExtra::Extra(
+                                ExtraName::new(extra.clone())
+                                    .with_context(|| format!("invalid extra name: {extra}"))?,
+                            ),
+                        };
+                        dep.marker.and(MarkerTree::expression(new_extra));
                         self.requires_dist.push(dep);
                     }
                 }
             }
 
             if let Some(scripts) = &project.scripts {
-                self.scripts = scripts.clone();
+                self.scripts.clone_from(scripts);
             }
             if let Some(gui_scripts) = &project.gui_scripts {
-                self.gui_scripts = gui_scripts.clone();
+                self.gui_scripts.clone_from(gui_scripts);
             }
             if let Some(entry_points) = &project.entry_points {
                 // Raise error on ambiguous entry points: https://www.python.org/dev/peps/pep-0621/#entry-points
@@ -322,7 +423,7 @@ impl Metadata21 {
                 if entry_points.contains_key("gui_scripts") {
                     bail!("gui_scripts is not allowed in project.entry-points table");
                 }
-                self.entry_points = entry_points.clone();
+                self.entry_points.clone_from(entry_points);
             }
         }
         Ok(())
@@ -334,7 +435,7 @@ impl Metadata21 {
     pub fn from_cargo_toml(
         manifest_path: impl AsRef<Path>,
         cargo_metadata: &cargo_metadata::Metadata,
-    ) -> Result<Metadata21> {
+    ) -> Result<Metadata24> {
         let package = cargo_metadata
             .root_package()
             .context("Expected cargo to return metadata with root_package")?;
@@ -379,11 +480,24 @@ impl Metadata21 {
         };
         let name = package.name.clone();
         let mut project_url = IndexMap::new();
+        if let Some(homepage) = package.homepage.as_ref() {
+            project_url.insert("Homepage".to_string(), homepage.clone());
+        }
+        if let Some(documentation) = package.documentation.as_ref() {
+            project_url.insert("Documentation".to_string(), documentation.clone());
+        }
         if let Some(repository) = package.repository.as_ref() {
             project_url.insert("Source Code".to_string(), repository.clone());
         }
         let license_files = if let Some(license_file) = package.license_file.as_ref() {
-            vec![manifest_path.as_ref().join(license_file)]
+            let license_path = manifest_path.as_ref().join(license_file).normalize()?;
+            if !license_path.is_file() {
+                bail!(
+                    "license file `{license_file}` specified in `{}` is not a file",
+                    manifest_path.as_ref().display()
+                );
+            }
+            vec![license_path.into_path_buf()]
         } else {
             Vec::new()
         };
@@ -397,10 +511,10 @@ impl Metadata21 {
                 err
             )
         })?;
-        let metadata = Metadata21 {
-            // name, version and metadata_version are added through Metadata21::new()
+        let metadata = Metadata24 {
+            // name, version and metadata_version are added through metadata24::new()
             // Mapped from cargo metadata
-            summary: package.description.clone(),
+            summary: package.description.as_ref().map(|d| d.trim().into()),
             description,
             description_content_type,
             keywords: if package.keywords.is_empty() {
@@ -420,7 +534,7 @@ impl Metadata21 {
             license: package.license.clone(),
             license_files,
             project_url,
-            ..Metadata21::new(name, version)
+            ..Metadata24::new(name, version)
         };
         Ok(metadata)
     }
@@ -460,7 +574,8 @@ impl Metadata21 {
         let license_files: Vec<String> = self
             .license_files
             .iter()
-            .map(|path| path.file_name().unwrap().to_str().unwrap().to_string())
+            // Use a portable path to ensure the metadata is consistent between Unix and Windows.
+            .map(|path| path.display().to_string().replace("\\", "/"))
             .collect();
         add_vec("License-File", &license_files);
 
@@ -478,6 +593,9 @@ impl Metadata21 {
         add_option("Author-email", &self.author_email);
         add_option("Maintainer", &self.maintainer);
         add_option("Maintainer-email", &self.maintainer_email);
+        // PEP 639
+        add_option("License-Expression", &self.license_expression);
+        // Deprecated by PEP 639
         add_option("License", &self.license.as_deref().map(fold_header));
         add_option(
             "Requires-Python",
@@ -553,6 +671,32 @@ impl Metadata21 {
             &self.get_version_escaped()
         ))
     }
+
+    /// Returns the name of the .data directory as defined in the wheel specification
+    pub fn get_data_dir(&self) -> PathBuf {
+        PathBuf::from(format!(
+            "{}-{}.data",
+            &self.get_distribution_escaped(),
+            &self.get_version_escaped()
+        ))
+    }
+}
+
+/// Escape email addresses with display name if necessary
+/// according to RFC 822 Section 3.3. "specials".
+fn escape_email_with_display_name(display_name: &str, email: &str) -> String {
+    if display_name.chars().any(|c| {
+        matches!(
+            c,
+            '(' | ')' | '<' | '>' | '@' | ',' | ';' | ':' | '\\' | '"' | '.' | '[' | ']'
+        )
+    }) {
+        return format!(
+            "\"{}\" <{email}>",
+            display_name.replace('\\', "\\\\").replace('\"', "\\\"")
+        );
+    }
+    format!("{display_name} <{email}>")
 }
 
 /// Fold long header field according to RFC 5322 section 2.2.3
@@ -579,18 +723,19 @@ fn fold_header(text: &str) -> String {
 }
 
 #[cfg(test)]
-mod test {
+mod tests {
     use super::*;
     use cargo_metadata::MetadataCommand;
     use expect_test::{expect, Expect};
     use indoc::indoc;
     use pretty_assertions::assert_eq;
+    use tempfile::TempDir;
 
     fn assert_metadata_from_cargo_toml(
         readme: &str,
         cargo_toml: &str,
         expected: Expect,
-    ) -> Metadata21 {
+    ) -> Metadata24 {
         let crate_dir = tempfile::tempdir().unwrap();
         let crate_path = crate_dir.path();
         let manifest_path = crate_path.join("Cargo.toml");
@@ -614,7 +759,7 @@ mod test {
             .exec()
             .unwrap();
 
-        let metadata = Metadata21::from_cargo_toml(crate_path, &cargo_metadata).unwrap();
+        let metadata = Metadata24::from_cargo_toml(crate_path, &cargo_metadata).unwrap();
 
         let actual = metadata.to_file_contents().unwrap();
 
@@ -646,7 +791,9 @@ mod test {
             authors = ["konstin <konstin@mailbox.org>"]
             name = "info-project"
             version = "0.1.0"
-            description = "A test project"
+            description = """
+A test project
+            """
             homepage = "https://example.org"
             readme = "REPLACE_README_PATH"
             keywords = ["ffi", "test"]
@@ -658,7 +805,7 @@ mod test {
         );
 
         let expected = expect![[r#"
-            Metadata-Version: 2.1
+            Metadata-Version: 2.4
             Name: info-project
             Version: 0.1.0
             Summary: A test project
@@ -667,6 +814,7 @@ mod test {
             Author: konstin <konstin@mailbox.org>
             Author-email: konstin <konstin@mailbox.org>
             Description-Content-Type: text/markdown; charset=UTF-8; variant=GFM
+            Project-URL: Homepage, https://example.org
 
             # Some test package
 
@@ -706,7 +854,7 @@ mod test {
             .manifest_path(manifest_dir.join("Cargo.toml"))
             .exec()
             .unwrap();
-        let mut metadata = Metadata21::from_cargo_toml(&manifest_dir, &cargo_metadata).unwrap();
+        let mut metadata = Metadata24::from_cargo_toml(&manifest_dir, &cargo_metadata).unwrap();
         let pyproject_toml = PyProjectToml::new(manifest_dir.join("pyproject.toml")).unwrap();
         metadata
             .merge_pyproject_toml(&manifest_dir, &pyproject_toml)
@@ -755,7 +903,7 @@ mod test {
             .manifest_path(manifest_dir.join("Cargo.toml"))
             .exec()
             .unwrap();
-        let mut metadata = Metadata21::from_cargo_toml(&manifest_dir, &cargo_metadata).unwrap();
+        let mut metadata = Metadata24::from_cargo_toml(&manifest_dir, &cargo_metadata).unwrap();
         let pyproject_toml = PyProjectToml::new(manifest_dir.join("pyproject.toml")).unwrap();
         metadata
             .merge_pyproject_toml(&manifest_dir, &pyproject_toml)
@@ -776,7 +924,7 @@ mod test {
             .manifest_path(manifest_dir.join("Cargo.toml"))
             .exec()
             .unwrap();
-        let metadata = Metadata21::from_cargo_toml(&manifest_dir, &cargo_metadata).unwrap();
+        let metadata = Metadata24::from_cargo_toml(&manifest_dir, &cargo_metadata).unwrap();
         assert!(metadata.description.unwrap().starts_with("# pyo3-mixed"));
         assert_eq!(
             metadata.description_content_type.unwrap(),
@@ -784,6 +932,23 @@ mod test {
         );
     }
 
+    #[test]
+    fn test_pep639() {
+        let manifest_dir = PathBuf::from("test-crates").join("pyo3-mixed");
+        let cargo_metadata = MetadataCommand::new()
+            .manifest_path(manifest_dir.join("Cargo.toml"))
+            .exec()
+            .unwrap();
+        let mut metadata = Metadata24::from_cargo_toml(&manifest_dir, &cargo_metadata).unwrap();
+        let pyproject_toml = PyProjectToml::new(manifest_dir.join("pyproject.toml")).unwrap();
+        metadata
+            .merge_pyproject_toml(&manifest_dir, &pyproject_toml)
+            .unwrap();
+
+        assert_eq!(metadata.license_expression.as_ref().unwrap(), "MIT");
+        assert_eq!(metadata.license.as_ref(), None);
+    }
+
     #[test]
     fn test_merge_metadata_from_pyproject_dynamic_license_test() {
         let manifest_dir = PathBuf::from("test-crates").join("license-test");
@@ -791,7 +956,7 @@ mod test {
             .manifest_path(manifest_dir.join("Cargo.toml"))
             .exec()
             .unwrap();
-        let mut metadata = Metadata21::from_cargo_toml(&manifest_dir, &cargo_metadata).unwrap();
+        let mut metadata = Metadata24::from_cargo_toml(&manifest_dir, &cargo_metadata).unwrap();
         let pyproject_toml = PyProjectToml::new(manifest_dir.join("pyproject.toml")).unwrap();
         metadata
             .merge_pyproject_toml(&manifest_dir, &pyproject_toml)
@@ -804,11 +969,195 @@ mod test {
         assert_eq!(4, metadata.license_files.len());
 
         // Verify pyproject.toml license = {file = ...} worked
-        assert_eq!(metadata.license_files[0], manifest_dir.join("LICENCE.txt"));
+        assert_eq!(metadata.license_files[0], PathBuf::from("LICENCE.txt"));
 
         // Verify the default licenses were included
-        assert_eq!(metadata.license_files[1], manifest_dir.join("LICENSE"));
-        assert_eq!(metadata.license_files[2], manifest_dir.join("NOTICE.md"));
-        assert_eq!(metadata.license_files[3], manifest_dir.join("AUTHORS.txt"));
+        assert_eq!(metadata.license_files[1], PathBuf::from("LICENSE"));
+        assert_eq!(metadata.license_files[2], PathBuf::from("NOTICE.md"));
+        assert_eq!(metadata.license_files[3], PathBuf::from("AUTHORS.txt"));
+    }
+
+    #[test]
+    fn test_escape_email_with_display_name_without_special_characters() {
+        let display_name = "Foo Bar !#$%&'*+-/=?^_`{|}~ 123";
+        let email = "foobar-123@example.com";
+        let result = escape_email_with_display_name(display_name, email);
+        assert_eq!(
+            result,
+            "Foo Bar !#$%&'*+-/=?^_`{|}~ 123 <foobar-123@example.com>"
+        );
+    }
+
+    #[test]
+    fn test_escape_email_with_display_name_with_special_characters() {
+        let tests = [
+            ("Foo ( Bar", "\"Foo ( Bar\""),
+            ("Foo ) Bar", "\"Foo ) Bar\""),
+            ("Foo < Bar", "\"Foo < Bar\""),
+            ("Foo > Bar", "\"Foo > Bar\""),
+            ("Foo @ Bar", "\"Foo @ Bar\""),
+            ("Foo , Bar", "\"Foo , Bar\""),
+            ("Foo ; Bar", "\"Foo ; Bar\""),
+            ("Foo : Bar", "\"Foo : Bar\""),
+            ("Foo \\ Bar", "\"Foo \\\\ Bar\""),
+            ("Foo \" Bar", "\"Foo \\\" Bar\""),
+            ("Foo . Bar", "\"Foo . Bar\""),
+            ("Foo [ Bar", "\"Foo [ Bar\""),
+            ("Foo ] Bar", "\"Foo ] Bar\""),
+            ("Foo ) Bar", "\"Foo ) Bar\""),
+            ("Foo ) Bar", "\"Foo ) Bar\""),
+            ("Foo, Bar", "\"Foo, Bar\""),
+        ];
+        for (display_name, expected_name) in tests {
+            let email = "foobar-123@example.com";
+            let result = escape_email_with_display_name(display_name, email);
+            let expected = format!("{expected_name} <{email}>");
+            assert_eq!(result, expected);
+        }
+    }
+
+    #[test]
+    fn test_issue_2544_respect_pyproject_dynamic_without_dynamic_fields() {
+        let temp_dir = TempDir::new().unwrap();
+        let crate_path = temp_dir.path();
+        let manifest_path = crate_path.join("Cargo.toml");
+        let pyproject_path = crate_path.join("pyproject.toml");
+
+        // Create basic src structure
+        fs::create_dir(crate_path.join("src")).unwrap();
+        fs::write(crate_path.join("src/lib.rs"), "").unwrap();
+
+        // Write Cargo.toml with metadata that should NOT be included
+        // because pyproject.toml doesn't declare them as dynamic
+        let cargo_toml = indoc!(
+            r#"
+            [package]
+            name = "test-package"
+            version = "0.1.0"
+            description = "Description from Cargo.toml - should not appear"
+            authors = ["author from cargo.toml <author@example.com>"]
+            keywords = ["cargo", "toml", "keyword"]
+            repository = "https://github.com/example/repo"
+
+            [lib]
+            crate-type = ["cdylib"]
+            "#
+        );
+        fs::write(&manifest_path, cargo_toml).unwrap();
+
+        // Write pyproject.toml WITHOUT declaring the fields as dynamic
+        let pyproject_toml_content = indoc!(
+            r#"
+            [build-system]
+            requires = ["maturin>=1.0,<2.0"]
+            build-backend = "maturin"
+
+            [project]
+            name = "test-package"
+            version = "0.1.0"
+            # Note: no description, authors, keywords, urls in dynamic list
+            # dynamic = []  # Not specified, so defaults to empty
+            "#
+        );
+        fs::write(&pyproject_path, pyproject_toml_content).unwrap();
+
+        // Load metadata as maturin does
+        let cargo_metadata = MetadataCommand::new()
+            .manifest_path(&manifest_path)
+            .exec()
+            .unwrap();
+        let mut metadata = Metadata24::from_cargo_toml(crate_path, &cargo_metadata).unwrap();
+        let pyproject_toml = PyProjectToml::new(&pyproject_path).unwrap();
+        metadata
+            .merge_pyproject_toml(crate_path, &pyproject_toml)
+            .unwrap();
+
+        assert_eq!(
+            metadata.summary, None,
+            "summary should be None when not in dynamic list"
+        );
+        assert_eq!(
+            metadata.author, None,
+            "author should be None when not in dynamic list"
+        );
+        assert_eq!(
+            metadata.keywords, None,
+            "keywords should be None when not in dynamic list"
+        );
+        assert!(
+            metadata.project_url.is_empty(),
+            "project_url should be empty when not in dynamic list"
+        );
+    }
+
+    #[test]
+    fn test_issue_2544_respect_pyproject_dynamic_with_dynamic_fields() {
+        let temp_dir = TempDir::new().unwrap();
+        let crate_path = temp_dir.path();
+        let manifest_path = crate_path.join("Cargo.toml");
+        let pyproject_path = crate_path.join("pyproject.toml");
+
+        // Create basic src structure
+        fs::create_dir(crate_path.join("src")).unwrap();
+        fs::write(crate_path.join("src/lib.rs"), "").unwrap();
+
+        // Write Cargo.toml with metadata
+        let cargo_toml = indoc!(
+            r#"
+            [package]
+            name = "test-package"
+            version = "0.1.0"
+            description = "Description from Cargo.toml - should appear"
+            authors = ["author from cargo.toml <author@example.com>"]
+            keywords = ["cargo", "toml", "keyword"]
+            repository = "https://github.com/example/repo"
+
+            [lib]
+            crate-type = ["cdylib"]
+            "#
+        );
+        fs::write(&manifest_path, cargo_toml).unwrap();
+
+        // Write pyproject.toml WITH fields declared as dynamic
+        let pyproject_toml_content = indoc!(
+            r#"
+            [build-system]
+            requires = ["maturin>=1.0,<2.0"]
+            build-backend = "maturin"
+
+            [project]
+            name = "test-package"
+            version = "0.1.0"
+            # Fields declared as dynamic - should come from Cargo.toml
+            dynamic = ["description", "authors", "keywords", "urls"]
+            "#
+        );
+        fs::write(&pyproject_path, pyproject_toml_content).unwrap();
+
+        // Load metadata as maturin does
+        let cargo_metadata = MetadataCommand::new()
+            .manifest_path(&manifest_path)
+            .exec()
+            .unwrap();
+        let mut metadata = Metadata24::from_cargo_toml(crate_path, &cargo_metadata).unwrap();
+        let pyproject_toml = PyProjectToml::new(&pyproject_path).unwrap();
+        metadata
+            .merge_pyproject_toml(crate_path, &pyproject_toml)
+            .unwrap();
+
+        // These fields SHOULD be set because they are in dynamic list
+        assert_eq!(
+            metadata.summary,
+            Some("Description from Cargo.toml - should appear".to_string())
+        );
+        assert_eq!(
+            metadata.author,
+            Some("author from cargo.toml <author@example.com>".to_string())
+        );
+        assert_eq!(metadata.keywords, Some("cargo,toml,keyword".to_string()));
+        assert_eq!(
+            metadata.project_url.get("Source Code"),
+            Some(&"https://github.com/example/repo".to_string())
+        );
     }
 }
